/*
--          This file is part of the New World OS and Objectify projects
--                     Copyright (C) 2009, 2010  QRW Software
--               J. Scott Edwards - j.scott.edwards.nwos@gmail.com 
--
--   This program is free software: you can redistribute it and/or modify
--   it under the terms of the GNU General Public License as published by
--   the Free Software Foundation, either version 3 of the License, or
--   (at your option) any later version.
--
--   This program is distributed in the hope that it will be useful,
--   but WITHOUT ANY WARRANTY; without even the implied warranty of
--   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--   GNU General Public License for more details.
--
--   You should have received a copy of the GNU General Public License
--   along with this program, in the file LICENSE.  If not, see 
--   <http://www.gnu.org/licenses/>.
--
--   For the latest information, source code (SVN), releases, and bug tracking
--   go to:
--      http://savannah.nongnu.org/projects/objectify
--
--   For releases from Alpha_30 and up, bug and feature request tracking go to:
--      http://sourceforge.net/projects/objectify
--
--   For older bug tracking, releases and source code (CVS) prior to the
--   Alpha_30 release go to:
--      http://sourceforge.net/projects/nwos
--
--   Other related websites:
--      http://www.qrwsoftware.com
--      http://www.worldwide-database.org
--
--   You can also contact me via paper mail at:
--
--      QRW Software
--      P.O. Box 27511
--      Salt Lake City, UT 84127-0511, USA.
--
--   $Author: jsedwards $
--   $Date: 2011-12-10 16:10:59 -0700 (Sat, 10 Dec 2011) $
--   $Revision: 4993 $
--
*/


#include <assert.h>
#include <limits.h>    /* define PATH_MAX */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "class_definition.h"
#include "header.h"
#include "objectify.h"
#include "storage.h"
#include "time_stamp.h"


static void display_usage(char* prog_name)
{
    fprintf(stderr, "usage: %s\n", prog_name);
    fprintf(stderr, "   or: %s --compressed file\n", prog_name);
    fprintf(stderr, "\n");
    fprintf(stderr, "  --compressed file\tuse file for archive.\n");
    fprintf(stderr, "\n");
}


static size_t get_path_object_size(void* file_path_obj)
{
    assert(((C_struct_File_Path*)file_path_obj)->count > 0);

    return sizeof(C_struct_File_Path) + ((C_struct_File_Path*)file_path_obj)->count;
}


#define MAX_REFS 1048576

int main(int argc, char* argv[])
{
    char* path = DEFAULT_FILE;
    ObjRef file_class_refs[NUM_FILE_CLASS_DEFINITIONS];
    ObjRef file_class_ref;
    ObjRef assoc_class_ref;
    ObjRef object_class;
    ObjRef real_object_class;
    ObjRef assoc_ref;
    C_struct_Class_Definition class_def_obj;
    C_struct_File file_obj;
    C_struct_Path_And_File_Association assoc_obj;
    C_struct_MD5sum md5_obj;
    C_struct_SHA1sum sha1_obj;
    C_struct_SHA256sum sha256_obj;
    C_struct_SHA512sum sha512_obj;
    uint8 kludge[MAX_PATH_OBJ_SIZE];
    C_struct_File_Path* ptr_path_obj = (C_struct_File_Path*)kludge;
    ReferenceList* ref_list;
    ReferenceList* file_ref_list;
    ReferenceList* temp_ref_list;
    int num_refs;
    uint64 num_blocks;                       /* NEED TO FIX THIS MESS!!! */
    int num_file_refs;
    int num_temp_refs;
    int real_type;
    int def;
    int i;
    int j;
    int k;
    char name[PATH_MAX];
    uint64 file_size;
    int num_files = 0;
    int num_failures = 0;
    Disk_Header disk_header;
    ObjRef* zone_refs;
    bool result = true;

    printf("\n");

//    if (argc == 3 && strcmp(argv[1], "--compressed") == 0)
//    if (argc == 3)
//    {
//	path = argv[2];
//    }
    /*    else */ if (argc != 2)
    {
	display_usage(argv[0]);
	exit(1);
    }

    nwos_log_arguments(0, NULL);

    nwos_save_refs("/tmp/remove_block_list.ref");    /* have to call this before initialize so it catches the root object */

    nwos_initialize_objectify(READ_ONLY, path);

    /* read all of the class_definition references */

    nwos_read_object_headers_from_disk(&nwos_private_class_definition_class_ref, &class_def_obj.header);

    ref_list = nwos_malloc_reference_list(&class_def_obj.header.object.references);

    num_refs = ref_list->common_header.num_refs;

    printf("class object: %08x  refs: %08x %d\n", nwos_ref_to_word(&nwos_private_class_definition_class_ref), nwos_ref_to_word(&class_def_obj.header.object.references), num_refs);

    for (i = 0; i < num_refs; i++)
    {
	if (!is_same_object(&ref_list->references[i], &nwos_private_class_definition_class_ref))
	{
	    nwos_read_object_headers_from_disk(&ref_list->references[i], &class_def_obj.header);

	    temp_ref_list = nwos_malloc_reference_list(&class_def_obj.header.object.references);

	    num_temp_refs = temp_ref_list->common_header.num_refs;

	    nwos_free_reference_list(temp_ref_list);

	    printf("class object: %08x  refs: %08x %d\n", nwos_ref_to_word(&ref_list->references[i]), nwos_ref_to_word(&class_def_obj.header.object.references), num_temp_refs);
	}
    }

    nwos_free_reference_list(ref_list);


    if (!nwos_find_private_class_definition("FILE", &file_class_ref))
    {
	printf("No files are stored in the system!\n\n");
	exit(1);
    }

    if (!nwos_find_private_class_definition("PATH AND FILE ASSOCIATION", &assoc_class_ref))
    {
	printf("NOTE: No path and file associations are stored in the system!\n\n");
    }

    assert(NUM_FILE_CLASS_DEFINITIONS == 3);

    copy_reference(&file_class_refs[0], nwos_get_file_001_reference());
    copy_reference(&file_class_refs[1], nwos_get_file_002_reference());
    copy_reference(&file_class_refs[2], nwos_get_file_class_reference());

    assert(is_same_object(&file_class_ref, &file_class_refs[2]));

    for (def = 0; def < NUM_FILE_CLASS_DEFINITIONS; def++)
    {
	printf("File class %03d reference: %08x\n\n", def + 1, nwos_ref_to_word(&file_class_refs[def]));

	if (nwos_is_public_reference(&file_class_refs[def]))
	{
	    printf("  NO private files of file class %03d exist\n", def + 1);
	}
	else
	{
	    assert(nwos_is_private_reference(&file_class_refs[def]));

	    nwos_read_object_headers_from_disk(&file_class_refs[def], &class_def_obj.header);

	    ref_list = nwos_malloc_reference_list(&class_def_obj.header.object.references);

	    num_refs = ref_list->common_header.num_refs;

	    printf("  num_refs: %d\n\n", num_refs);

	    for (i = 0; i < num_refs; i++)
	    {
		nwos_get_object_class(&ref_list->references[i], &object_class);

		if (is_same_object(&object_class, &file_class_ref))
		{
		    num_files++;

		    printf("file: %d\n", num_files);

		    /* verify it is in the correct list */
		    nwos_get_object_class_without_update(&ref_list->references[i], &real_object_class);

		    if (!is_same_object(&real_object_class, &file_class_refs[def]))
		    {
			real_type = 0;

			if (is_same_object(&real_object_class, &file_class_refs[0]))
			{
			    real_type = 1;
			}
			else if (is_same_object(&real_object_class, &file_class_refs[1]))
			{
			    real_type = 2;
			}
			else if (is_same_object(&real_object_class, &file_class_refs[2]))
			{
			    real_type = 3;
			}

			assert(1 <= real_type && real_type <= 3);

			printf("Warning: file object is in reference list of %03d but is of type %03d\n", def + 1, real_type);
		    }

		    nwos_read_object_from_disk(&ref_list->references[i], &file_obj, sizeof(file_obj));

		    file_size = ((uint64)file_obj.size[0] << 32) |
				((uint64)file_obj.size[1] << 24) |
				((uint64)file_obj.size[2] << 16) |
				((uint64)file_obj.size[3] << 8)  |
				 (uint64)file_obj.size[4];

		    printf("         id: %08x\n", nwos_ref_to_word(&file_obj.header.common.id));
		    printf("       size: %llu\n", file_size);

		    num_temp_refs = 0;
		    if (!is_void_reference(&file_obj.md5sum))
		    {
			nwos_read_object_from_disk(&file_obj.md5sum, &md5_obj, sizeof(md5_obj));
			temp_ref_list = nwos_malloc_reference_list(&md5_obj.header.object.references);
			num_temp_refs = temp_ref_list->common_header.num_refs;
			nwos_free_reference_list(temp_ref_list);
		    }
		    printf("     md5sum: %08x  refs: %08x %d\n", nwos_ref_to_word(&file_obj.md5sum), nwos_ref_to_word(&md5_obj.header.object.references), num_temp_refs);

		    num_temp_refs = 0;
		    if (!is_void_reference(&file_obj.sha1sum))
		    {
			nwos_read_object_from_disk(&file_obj.sha1sum, &sha1_obj, sizeof(sha1_obj));
			temp_ref_list = nwos_malloc_reference_list(&sha1_obj.header.object.references);
			num_temp_refs = temp_ref_list->common_header.num_refs;
			nwos_free_reference_list(temp_ref_list);
		    }
		    printf("    sha1sum: %08x  refs: %08x %d\n", nwos_ref_to_word(&file_obj.sha1sum), nwos_ref_to_word(&sha1_obj.header.object.references), num_temp_refs);

		    num_temp_refs = 0;
		    if (!is_void_reference(&file_obj.sha256sum))
		    {
			nwos_read_object_from_disk(&file_obj.sha256sum, &sha256_obj, sizeof(sha256_obj));
			temp_ref_list = nwos_malloc_reference_list(&sha256_obj.header.object.references);
			num_temp_refs = temp_ref_list->common_header.num_refs;
			nwos_free_reference_list(temp_ref_list);
		    }
		    printf("  sha256sum: %08x  refs: %08x %d\n", nwos_ref_to_word(&file_obj.sha256sum), nwos_ref_to_word(&sha256_obj.header.object.references), num_temp_refs);

		    num_temp_refs = 0;
		    if (!is_void_reference(&file_obj.sha512sum))
		    {
			nwos_read_object_from_disk(&file_obj.sha512sum, &sha512_obj, sizeof(sha512_obj));
			temp_ref_list = nwos_malloc_reference_list(&sha512_obj.header.object.references);
			num_temp_refs = temp_ref_list->common_header.num_refs;
			nwos_free_reference_list(temp_ref_list);
		    }
		    printf("  sha512sum: %08x  refs: %08x %d\n", nwos_ref_to_word(&file_obj.sha512sum), nwos_ref_to_word(&sha512_obj.header.object.references), num_temp_refs);
//    ObjRef media;
		    printf(" block_list: %08x\n", nwos_ref_to_word(&file_obj.block_list));

		    if (!is_void_reference(&assoc_class_ref))
		    {
			file_ref_list = nwos_malloc_reference_list(&file_obj.header.object.references);

			num_file_refs = file_ref_list->common_header.num_refs;

			void_reference(&assoc_ref);   /* no association yet */

			for (j = 0; j < num_file_refs; j++)
			{
			    nwos_get_object_class(&file_ref_list->references[j], &object_class);

			    if (is_same_object(&object_class, &assoc_class_ref))
			    {
				nwos_read_object_from_disk(&file_ref_list->references[j], &assoc_obj, sizeof(assoc_obj));

				temp_ref_list = nwos_malloc_reference_list(&assoc_obj.header.object.references);
				num_temp_refs = temp_ref_list->common_header.num_refs;
				nwos_free_reference_list(temp_ref_list);

				printf("      assoc: %08x  refs: %08x %d\n", nwos_ref_to_word(&file_ref_list->references[j]), nwos_ref_to_word(&assoc_obj.header.object.references), num_temp_refs);

				printf("          time: %s\n", nwos_time_stamp_to_string(assoc_obj.modification_time));

				assert(nwos_read_variable_sized_object_from_disk(&assoc_obj.path, kludge, sizeof(kludge), &get_path_object_size));

				/* remember ptr_path_obj points to the kludge buffer */

				for (k = 0; k < ptr_path_obj->count; k++) name[k] = ptr_path_obj->storage[k];

				name[k] = '\0';

				temp_ref_list = nwos_malloc_reference_list(&ptr_path_obj->header.object.references);
				num_temp_refs = temp_ref_list->common_header.num_refs;
				nwos_free_reference_list(temp_ref_list);

				printf("          path: %s   refs: %08x %d\n", name, nwos_ref_to_word(&ptr_path_obj->header.object.references), num_temp_refs);

				if (is_same_object(&assoc_obj.file, &file_obj.header.common.id))
				{
				    copy_reference(&assoc_ref, &file_ref_list->references[j]);
				}
				else
				{
				    printf("   ERROR: association object points to a different file: %08x\n", nwos_ref_to_word(&assoc_obj.file));
				    num_failures++;
				    result = false;
				}
			    }
			}

			if (is_void_reference(&assoc_ref))
			{
			    printf("   ERROR: no association objects found for this file!\n");
			    num_failures++;
			    result = false;
			}
			else if (is_void_reference(&file_obj.block_list))
			{
			    printf("  File data NOT stored!\n");
			}
#ifdef DONT_DO_FILE_DATA
			else
			{
			    if (!nwos_restore_file(&assoc_ref, NULL))     /* calling restore_file with no file name just verfies the file */
			    {
				num_failures++;
			    }
			}
#endif
			nwos_free_reference_list(file_ref_list);
			file_ref_list = NULL;
		    }

		    printf("\n");
		}
	    }    

	    nwos_free_reference_list(ref_list);
	    ref_list = NULL;
	}

	printf("\n");
    }

    nwos_save_refs(NULL);  /* stop storing the refs */

    printf("Number of files checked: %d\n", num_files);
    printf("Number of files failed:  %d\n", num_failures);

    if (num_failures == 0)
    {
	ObjRef last_ref;
	FILE* remove_refs_fp;
	FILE* compressed_fp;
	uint8 block[FILE_BLOCK_SIZE];

	remove_refs_fp = fopen("/tmp/remove_block_list.ref", "r");

	if (remove_refs_fp == NULL)
	{
	    perror("open /tmp/remove_block_list.ref");
	    exit(1);
	}

	zone_refs = nwos_malloc(sizeof(ObjRef) * MAX_REFS);

	num_refs = 0;

	while (fread(&last_ref, 1, 4, remove_refs_fp) == 4)
	{
	    assert(nwos_read_block(&last_ref, block));

//	      printf("%08x\n", nwos_ref_to_word(&last_ref));

	    for (i = 0; i < num_refs; i++)
	    {
		if (is_same_object(&last_ref, &zone_refs[i]))
		{
		    i = num_refs + 1;
		    break;
		}

		assert(num_refs < MAX_REFS);

		if (nwos_ref_to_word(&last_ref) < nwos_ref_to_word(&zone_refs[i]))
		{
		    for (j = num_refs; j > i; j--) copy_reference(&zone_refs[j], &zone_refs[j-1]);
		    break;
		}
	    }

	    if (i <= num_refs)
	    {
		copy_reference(&zone_refs[i], &last_ref);
		num_refs++;
	    }

	    if (num_refs == MAX_REFS)
	    {
		fprintf(stderr, "This version can only handle %d refs.\n", MAX_REFS);
		nwos_terminate_objectify();
		exit(1);
	    }
	}

	if (ferror(remove_refs_fp))
	{
	    perror("read /tmp/remove_block_list.ref");
	    exit(1);
	}

	if (fclose(remove_refs_fp) != 0)
	{
	    perror("close /tmp/remove_block_list.ref");
	}

	compressed_fp = fopen(argv[1], "w");

	if (compressed_fp == NULL)
	{
	    perror(argv[1]);
	    nwos_terminate_objectify();
	    exit(1);
	}

	memset(block, 0, sizeof(block));

	if (fwrite(block, 1, sizeof(block), compressed_fp) != sizeof(block))
	{
	    perror(argv[1]);
	    nwos_terminate_objectify();
	    exit(1);
	}

	for (i = 0; i < num_refs; i++)
	{
	    if (!nwos_read_block(&zone_refs[i], block))
	    {
		fprintf(stderr, "Error reading block %d: %08x\n", i, nwos_ref_to_word(&zone_refs[i]));
		fclose(compressed_fp);
		nwos_terminate_objectify();
		exit(1);
	    }

	    if (fwrite(block, 1, sizeof(block), compressed_fp) != sizeof(block))
	    {
		perror(argv[1]);
		fclose(compressed_fp);
		nwos_terminate_objectify();
		exit(1);
	    }
	}

	if (fseek(compressed_fp, 0, SEEK_SET) < 0)
	{
	    perror(argv[1]);
	    fclose(compressed_fp);
	    nwos_terminate_objectify();
	    exit(1);
	}

	memcpy(disk_header.magic_number,   MAGIC_NUMBER,         sizeof(disk_header.magic_number));
	memcpy(disk_header.version_string, nwos_version_string,  sizeof(disk_header.version_string));
	memcpy(disk_header.type_code,      TYPE_CODE_COMPRESSED, sizeof(disk_header.type_code));
	nwos_uint32_to_4_uint8(&nwos_block_offset_to_chunks, disk_header.block_offset_to_chunks);

	nwos_get_time_stamp(disk_header.last_prep_disk);
	nwos_uint32_to_4_uint8(&nwos_total_private_chunks, disk_header.total_chunks);
	nwos_uint32_to_4_uint8(&nwos_used_private_chunks,  disk_header.used_chunks);

	num_blocks = num_refs;
	nwos_uint64_to_8_uint8(&num_blocks,  disk_header.used_blocks);

	nwos_time_stamp_copy(disk_header.last_change, nwos_last_change);

	if (fwrite(&disk_header, 1, sizeof(disk_header), compressed_fp) != sizeof(disk_header))
	{
	    perror(argv[1]);
	    fclose(compressed_fp);
	    nwos_terminate_objectify();
	    exit(1);
	}

	if (fclose(compressed_fp) != 0)
	{
	    perror(argv[1]);
	    nwos_terminate_objectify();
	    exit(1);
	}
    }

    nwos_terminate_objectify();

    return !result;    /* return 0 means success */
}

